#!/usr/bin/python3

#
# This script configures a mimofb in NM mode to decrease the current
# of the strongest SR corrector by acting on the correctors before
# and after the strongest one.
# The optimizer keep constant the rms of the orbit
#
# Usage:
#	DischargeCorr.py [selfseqdev] [plane] 
#
# Example:
# 	./DischargeCorr.py seq/discharge/psch_sr hor
#
# 	./DischargeCorr.py seq/discharge/pscv_sr ver
#

import sys,re
from PyTango import *
import time
import numpy as np

DEBUG = 0

def main():

	max_val = 0
	target_val = 0
	pre_idx = 0
	post_idx = 0	
	max_idx = 0
	rms_ref = 0

	try:
		selfseq_dev = DeviceProxy(sys.argv[1])
		plane = sys.argv[2]
		
		# can't pass all device names to exec bacause of bash length limit
		if plane == "hor":
			mimofb_name = "opt/sr/discharge_psch_sr"
		else:
			mimofb_name = "opt/sr/discharge_pscv_sr"					

		mean_orbit_name = "sr/feedback/profile_s"

		mimo_dev = DeviceProxy(mimofb_name)
		mean_orbit_dev = DeviceProxy(mean_orbit_name)
		
		selfseq_dev.LogMsg = "Initialization done"
		
	except DevFailed as df:
		if DEBUG == 1:
			print("Error during deviceproxy creation,",str(df))
		selfseq_dev.LogMsg = "Error during deviceproxy creation: " + str(df)
		sys.exit(1)

	# this is the current in Amps the optimizer has to decrease on the
	# strongest corrector 
	discharge_curr = selfseq_dev.DischargeCurrent

	# create a ps group to read the actual current settings
	device_group = Group("group")
	if plane == "hor":
		device_group.add("sr/power_supply/psch_s1.*")
		device_group.add("sr/power_supply/psch_s2.*")
		device_group.add("sr/power_supply/psch_s3.*")
		device_group.add("sr/power_supply/psch_s4.*")
		device_group.add("sr/power_supply/psch_s5.*")
		device_group.add("sr/power_supply/psch_s6.1")
		device_group.add("sr/power_supply/psch_s6.2")
		device_group.add("sr/power_supply/psch_s6.8")
		device_group.add("sr/power_supply/psch_s6.3")
		device_group.add("sr/power_supply/psch_s6.*")								
		device_group.add("sr/power_supply/psch_s7.1")
		device_group.add("sr/power_supply/psch_s7.2")
		device_group.add("sr/power_supply/psch_s7.8")
		device_group.add("sr/power_supply/psch_s7.3")
		device_group.add("sr/power_supply/psch_s7.*")	
		device_group.add("sr/power_supply/psch_s8.1")
		device_group.add("sr/power_supply/psch_s8.2")
		device_group.add("sr/power_supply/psch_s8.8")
		device_group.add("sr/power_supply/psch_s8.3")
		device_group.add("sr/power_supply/psch_s8.*")
		device_group.add("sr/power_supply/psch_s9.1")
		device_group.add("sr/power_supply/psch_s9.2")
		device_group.add("sr/power_supply/psch_s9.8")
		device_group.add("sr/power_supply/psch_s9.3")
		device_group.add("sr/power_supply/psch_s9.*")
		device_group.add("sr/power_supply/psch_s10.1")
		device_group.add("sr/power_supply/psch_s10.2")
		device_group.add("sr/power_supply/psch_s10.8")
		device_group.add("sr/power_supply/psch_s10.3")
		device_group.add("sr/power_supply/psch_s10.*")																	
		device_group.add("sr/power_supply/psch_s11.1")
		device_group.add("sr/power_supply/psch_s11.2")
		device_group.add("sr/power_supply/psch_s11.8")
		device_group.add("sr/power_supply/psch_s11.3")
		device_group.add("sr/power_supply/psch_s11.*")	
		device_group.add("sr/power_supply/psch_s12.*")				

		if selfseq_dev.PreserveZigZag:
			device_group.remove("sr/power_supply/psch_s9.2")
			device_group.remove("sr/power_supply/psch_s9.3")
			device_group.remove("sr/power_supply/psch_s9.4")
			device_group.remove("sr/power_supply/psch_s9.5")									
	else:
		device_group.add("sr/power_supply/pscv_s1.*")
		device_group.add("sr/power_supply/pscv_s2.*")
		device_group.add("sr/power_supply/pscv_s3.*")
		device_group.add("sr/power_supply/pscv_s4.*")
		device_group.add("sr/power_supply/pscv_s5.*")
		device_group.add("sr/power_supply/pscv_s6.1")
		device_group.add("sr/power_supply/pscv_s6.2")
		device_group.add("sr/power_supply/pscv_s6.8")
		device_group.add("sr/power_supply/pscv_s6.3")
		device_group.add("sr/power_supply/pscv_s6.*")								
		device_group.add("sr/power_supply/pscv_s7.1")
		device_group.add("sr/power_supply/pscv_s7.2")
		device_group.add("sr/power_supply/pscv_s7.8")
		device_group.add("sr/power_supply/pscv_s7.3")
		device_group.add("sr/power_supply/pscv_s7.*")	
		device_group.add("sr/power_supply/pscv_s8.1")
		device_group.add("sr/power_supply/pscv_s8.2")
		device_group.add("sr/power_supply/pscv_s8.8")
		device_group.add("sr/power_supply/pscv_s8.3")
		device_group.add("sr/power_supply/pscv_s8.*")
		device_group.add("sr/power_supply/pscv_s9.1")
		device_group.add("sr/power_supply/pscv_s9.2")
		device_group.add("sr/power_supply/pscv_s9.8")
		device_group.add("sr/power_supply/pscv_s9.3")
		device_group.add("sr/power_supply/pscv_s9.*")
		device_group.add("sr/power_supply/pscv_s10.1")
		device_group.add("sr/power_supply/pscv_s10.2")
		device_group.add("sr/power_supply/pscv_s10.8")
		device_group.add("sr/power_supply/pscv_s10.3")
		device_group.add("sr/power_supply/pscv_s10.*")																	
		device_group.add("sr/power_supply/pscv_s11.1")
		device_group.add("sr/power_supply/pscv_s11.2")
		device_group.add("sr/power_supply/pscv_s11.8")
		device_group.add("sr/power_supply/pscv_s11.3")
		device_group.add("sr/power_supply/pscv_s11.*")	
		device_group.add("sr/power_supply/pscv_s12.*")	

	# this is the list of the power supplies
	try:	
		ps_list = device_group.get_device_list()

	except DevFailed as df:
		if DEBUG == 1:
			print("Error creating ps group list,",str(df))
		selfseq_dev.LogMsg = "Error creating ps group list," + str(df)
		sys.exit(1)	
	
	# now read currents
	try:
		reply=device_group.read_attribute('Current')	

	except DevFailed as df:
		if DEBUG == 1:
			print("Error reading ps currents,",str(df))
		selfseq_dev.LogMsg = "Error reading ps currents," + str(df)
		sys.exit(1)
				
	list_data = []
	
	max_val = 0
	target_val = 0	
	max_idx = 0
	
	# find the ps with the maximum settings
	for i in range(0,len(reply)-1):
		if reply[i].get_data().has_failed == False:
			print(ps_list[i],reply[i].get_data().value)
			list_data.append([ps_list[i],reply[i].get_data().value])
			if abs(reply[i].get_data().value) > max_val:
				max_val = abs(reply[i].get_data().value)
				max_idx = i	
				if reply[i].get_data().value > 0:
					target_val = reply[i].get_data().value - discharge_curr
				else:
					target_val = reply[i].get_data().value + discharge_curr													
	
	pre_idx = np.mod(max_idx-1,len(reply)-1)
	post_idx = np.mod(max_idx+1,len(reply)-1)		

	selfseq_dev.LogMsg = "Discharging corrector " + ps_list[max_idx] + ", new target current " + str(target_val)

	# read orbit rms
	try:
		if plane == "hor":
			rms_ref = mean_orbit_dev.AbsRmsOrbitHor
		else:
			rms_ref = mean_orbit_dev.AbsRmsOrbitVer	
	except DevFailed as df:
		if DEBUG == 1:
			print("Error reading orbit rms,",str(df))
		selfseq_dev.LogMsg = "Error reading orbit rms," + str(df)
		sys.exit(1)				

	# check if data is valid		
	if np.isnan(rms_ref):
		if DEBUG == 1:
			print("Error: Nan rms orbit value")
		selfseq_dev.LogMsg = "Error: Nan rms orbit value"
		sys.exit(1)	

	# check if data is valid		
	if rms_ref > 0.6:
		if DEBUG == 1:
			print("Error: rms orbit > 0.6")
		selfseq_dev.LogMsg = "Error: rms orbit > 0.6"
		sys.exit(1)


	# if mimofb is already running exit
	if (str(mimo_dev.State()) == 'ON'):
		if DEBUG == 1:
			print("Error: MIMOFB is already ON")
		selfseq_dev.LogMsg = "Error: MIMOFB is already ON"
		sys.exit(1)
	
	# actuator configuration
	config_actuators = []
	config_actuators.append(ps_list[pre_idx]+';Current;'+ps_list[pre_idx]+';1;-16;16;16;0.2;0;0;1;0;500;ON');
	config_actuators.append(ps_list[max_idx]+';Current;'+ps_list[max_idx]+';1;-16;16;16;0.2;0;0;1;0;500;ON');
	config_actuators.append(ps_list[post_idx]+';Current;'+ps_list[post_idx]+';1;-16;16;16;0.2;0;0;1;0;500;ON');	
	config_actuators.append('NONE;NONE');
	config_actuators.append('NONE;NONE');	
	config_actuators.append('NONE;NONE');
		
	# sensor configuration			
	config_sensors = []
	if plane == "hor":
		config_sensors.append(mean_orbit_name+';AbsRmsOrbitHor;AbsRmsOrbitHor;1;0;0.6;'+'{:03f}'.format(rms_ref)+';0;500;0;ON,OFF,STANDBY;0,1,2');	
	else:
		config_sensors.append(mean_orbit_name+';AbsRmsOrbitVer;AbsRmsOrbitVer;1;0;0.6;'+'{:03f}'.format(rms_ref)+';0;500;0;ON,OFF,STANDBY;0,1,2');		
	config_sensors.append(ps_list[max_idx]+';Current;'+ps_list[max_idx]+';5;-16;16;'+'{:03f}'.format(target_val)+';0;500;0;ON;0,1,2');
	config_sensors.append('NONE;NONE');
	config_sensors.append('NONE;NONE');
	
	if DEBUG:
		print(config_actuators)
		print(config_sensors)

	#dev_mimo.put_property({'ActuatorConfiguration':config_actuators})
	try:
		# write mimo configuration 
		mimo_dev.write_attribute('ConfigActuators',config_actuators)
		mimo_dev.write_attribute('ConfigSensors',config_sensors)
		
		# Nelder Mead
		mimo_dev.ActionMode = 1
		mimo_dev.CountdownMode = True
		mimo_dev.CountdownMaxTime = 100
		mimo_dev.LimitKickFactor = 1
		
		# turn on the optimizer
		mimo_dev.On()
		time.sleep(5)
		
		
		# check the mimo is running (ON)
		while (str(mimo_dev.State()) == 'ON'):
		
			# if sequencer is not running exit
			if DEBUG == 0:
				if (str(selfseq_dev.State()) != 'RUNNING'): 
					if DEBUG == 1:
						print("Error: sequencer is not RUNNING")
					selfseq_dev.LogMsg = "Error: sequencer is not RUNNING"
					sys.exit(1)	
		
			time.sleep(2)
		
		# turn off the MIMO	
		mimo_dev.Off()	
			
	except DevFailed as df:
		if DEBUG == 1:
			print("Failed to configure/run MIMOFB,",str(df))
		selfseq_dev.LogMsg = "Failed to configure/run MIMOFB: " + str(df)
		sys.exit(1)				
				
				
				
if __name__ == "__main__":		
	main()				
